//*********************************************************************
//                 Copyright (C) Microsoft Corporation.
//
// @File: umtracingapi.h
// @Owner: adsmiech
//
// Purpose:
//   A header file to accompany tracing.dll. 
//
//*********************************************************************
#pragma once
#ifndef UMTRACINGAPI_H
#define UMTRACINGAPI_H

#ifdef WIN32

#ifdef TRACINGAPI

// TRACINGAPI should be defined in all of the DLL's source
// code modules before this header file is included.

// All functions/variables are being exported.

#else
#define THIS_IS_CLIENT

// This header file is included by an EXE source code module.
// Indicate that all functions/variables are being imported.

#ifdef __cplusplus
#define TRACINGAPI extern "C" __declspec(dllimport)
#else
#define TRACINGAPI __declspec(dllimport)
#endif

#endif
#define _ALIGN(x)   __declspec(align(x))
////////////////////////////////////////////////////////////////////////////

// Define any data structures and symbols here.


// CAUTION:
// If you are going to change the current layout of the defined fields
// or the size of this structure, please make that dispatch.asm is adjusted
// accordingly.
//
_ALIGN(64)
typedef struct _KTR_ENTRY
{
    /// <summary>
    /// The timestamp of event happened
    /// </summary>
    ULONG64 Timestamp;

    /// <summary>
    /// The pointer of the c-style format string. this can be used as
    /// unique id for the trace. This format string has to be const.
    /// </summary>
    const CHAR* Description;

    /// <summary>
    /// The pointer of TEB.  This eases thread based KTR log analysis.
    /// </summary>
    PVOID Teb;

    /// <summary>
    /// One of the KTR_TRACE_FLAGS_*.  This eases identity based KTR filtering.
    /// </summary>
    ULONG64 Identity;

    /// <summary>
    /// At most six parameters to record. Only value is recorded.
    /// The first 4 arguments are in the same cacheline as KTR_ENTRY header.
    /// </summary>
    PVOID Arguments[6];
}KTR_ENTRY, *PKTR_ENTRY;

static_assert(sizeof(KTR_ENTRY) == 128, "Size of KTR_ENTRY os mpt 128");

// CAUTION:
// - KTR_ENTRY_MAX has to be 2^N.
//
#define KTR_ENTRY_MAX           0x80000
#define KTR_ENTRY_MASK          (KTR_ENTRY_MAX - 1)

static_assert((KTR_ENTRY_MAX & KTR_ENTRY_MASK) == 0, "(KTR_ENTRY_MAX & KTR_ENTRY_MASK) != 0");

typedef enum _KTR_TRACE_INDEX
{
    KTR_TRACE_INDEX_SYSCALL = 0,
    KTR_TRACE_INDEX_IO = 1,
    KTR_TRACE_INDEX_TIMER = 2,
    KTR_TRACE_INDEX_DKCALL = 3,
    KTR_TRACE_INDEX_INIT = 4,
    KTR_TRACE_INDEX_PROCESS = 5,
    KTR_TRACE_INDEX_THREAD = 6,
    KTR_TRACE_INDEX_SYNCH = 7,
    KTR_TRACE_INDEX_EXCEPTION = 8,
    // More indices here.
    KTR_TRACE_INDEX_MAX // not an index.
} KTR_TRACE_INDEX;

static_assert(KTR_TRACE_INDEX_MAX < 64, "We can only have up to 63 KTR_TRACE_INDICES"); // g_TraceMask is 64bits.

#define _KTR_TRACE_FLAG(x)      KTR_TRACE_FLAGS_##x = (1ULL << KTR_TRACE_INDEX_##x)
typedef enum _KTR_TRACE_FLAGS
{
    _KTR_TRACE_FLAG(SYSCALL),
    _KTR_TRACE_FLAG(IO),
    _KTR_TRACE_FLAG(TIMER),
    _KTR_TRACE_FLAG(DKCALL),
    _KTR_TRACE_FLAG(INIT),
    _KTR_TRACE_FLAG(PROCESS),
    _KTR_TRACE_FLAG(THREAD),
    _KTR_TRACE_FLAG(SYNCH),
    _KTR_TRACE_FLAG(EXCEPTION)
}KTR_TRACE_FLAGS;
#undef _KTR_TRACE_FLAG



typedef void(*pfTrace)(
    KTR_TRACE_FLAGS flags,
    const CHAR*     Description,
    PVOID           Arguments1,
    PVOID           Arguments2,
    PVOID           Arguments3,
    PVOID           Arguments4,
    PVOID           Arguments5,
    PVOID           Arguments6,
    INT             NumOfArugments
);

#ifdef KTR_INSTRUMENTATION

#define KTR0(flags, Description) \
    do \
    { \
        if (pTraceMask && (flags & (*pTraceMask)) != 0 ) \
        { \
            pfLogTrace(flags,\
                        Description,\
                        (PVOID)NULL,\
                        (PVOID)NULL,\
                        (PVOID)NULL,\
                        (PVOID)NULL,\
                        (PVOID)NULL,\
                        (PVOID)NULL,\
                        0); \
        } \
    } while(0)


#define KTR1(flags, Description, Arg1) \
    do \
    { \
        if (pTraceMask && (flags & (*pTraceMask)) != 0 ) \
        { \
            pfLogTrace(flags,\
                        Description,\
                        (PVOID)Arg1,\
                        (PVOID)NULL,\
                        (PVOID)NULL,\
                        (PVOID)NULL,\
                        (PVOID)NULL,\
                        (PVOID)NULL,\
                        1); \
        } \
    } while(0)


#define KTR2(flags, Description, Arg1, Arg2) \
    do \
    { \
        if (pTraceMask && (flags & (*pTraceMask)) != 0 ) \
        { \
            pfLogTrace(flags,\
                   Description,\
                   (PVOID)Arg1,\
                   (PVOID)Arg2,\
                   (PVOID)NULL,\
                   (PVOID)NULL,\
                   (PVOID)NULL,\
                   (PVOID)NULL,\
                   2); \
        } \
    } while(0)


#define KTR3(flags, Description, Arg1, Arg2, Arg3) \
    do \
    { \
        if (pTraceMask && (flags & (*pTraceMask)) != 0 ) \
        { \
            pfLogTrace(flags,\
                   Description,\
                   (PVOID)Arg1,\
                   (PVOID)Arg2,\
                   (PVOID)Arg3,\
                   (PVOID)NULL,\
                   (PVOID)NULL,\
                   (PVOID)NULL,\
                   3); \
        } \
    } while(0)


#define KTR4(flags, Description, Arg1, Arg2, Arg3, Arg4) \
    do \
    { \
        if (pTraceMask && (flags & (*pTraceMask)) != 0 ) \
        { \
            pfLogTrace(flags,\
                   Description,\
                   (PVOID)Arg1,\
                   (PVOID)Arg2,\
                   (PVOID)Arg3,\
                   (PVOID)Arg4,\
                   (PVOID)NULL,\
                   (PVOID)NULL,\
                   4); \
        } \
    } while(0)


#define KTR5(flags, Description, Arg1, Arg2, Arg3, Arg4, Arg5) \
    do \
    { \
        if (pTraceMask && (flags & (*pTraceMask)) != 0 ) \
        { \
            pfLogTrace(flags,\
                   Description,\
                   (PVOID)Arg1,\
                   (PVOID)Arg2,\
                   (PVOID)Arg3,\
                   (PVOID)Arg4,\
                   (PVOID)Arg5,\
                   (PVOID)NULL,\
                   5); \
        } \
    } while(0)


#define KTR6(flags, Description, Arg1, Arg2, Arg3, Arg4, Arg5, Arg6) \
    do { \
        if (pTraceMask && (flags & (*pTraceMask)) != 0 ) \
        { \
            pfLogTrace(flags,\
                   Description,\
                   (PVOID)Arg1,\
                   (PVOID)Arg2,\
                   (PVOID)Arg3,\
                   (PVOID)Arg4,\
                   (PVOID)Arg5,\
                   (PVOID)Arg6,\
                   6); \
        } \
    } while(0)

#else // KTR_INSTRUMENTATION

#define KTR0(flags, Description)
#define KTR1(flags, Description, Arg1)
#define KTR2(flags, Description, Arg1, Arg2)
#define KTR3(flags, Description, Arg1, Arg2, Arg3)
#define KTR4(flags, Description, Arg1, Arg2, Arg3, Arg4)
#define KTR5(flags, Description, Arg1, Arg2, Arg3, Arg4, Arg5)
#define KTR6(flags, Description, Arg1, Arg2, Arg3, Arg4, Arg5, Arg6)

#endif // KTR_INSTRUMENTATION

// Convenient function trace/probe macros.
//
// TRACE_{ENTER,LEAVE,PROBE}0
// TRACE_{ENTER,LEAVE,PROBE}1
// TRACE_{ENTER,LEAVE,PROBE}2
// TRACE_{ENTER,LEAVE,PROBE}3
// TRACE_{ENTER,LEAVE,PROBE}4
//
// - Currently 3 types of annotations are supported:
//   Func:enter, function 'Func' enters.
//   Func:leave, function 'Func' leaves.
//   Ident:probe, 'Ident' happens.
// - All arguments are logged as 0x%I64x. This prints a PVOID pointer as a hexadecimal
// - String arguments are _not_ supported.  Well, KTR was not designed
//   to log dynamic strings; while constant strings should be embedded
//   in 'Description'.
//

#define TRACE_EVENT0(ident, call, event) \
    KTR0(KTR_TRACE_FLAGS_##ident, #call ":" #event)
#define TRACE_ENTER0(ident, call) \
    TRACE_EVENT0(ident, call, enter)
#define TRACE_LEAVE0(ident, call) \
    TRACE_EVENT0(ident, call, leave)
#define TRACE_PROBE0(ident, call) \
    TRACE_EVENT0(ident, call, probe)

#define TRACE_EVENT1(ident, call, event, arg1) \
    KTR1(KTR_TRACE_FLAGS_##ident, \
         #call ":" #event \
         " " #arg1 "=0x%I64x", \
         arg1)
#define TRACE_ENTER1(ident, call, arg1) \
    TRACE_EVENT1(ident, call, enter, arg1)
#define TRACE_LEAVE1(ident, call, arg1) \
    TRACE_EVENT1(ident, call, leave, arg1)
#define TRACE_PROBE1(ident, call, arg1) \
    TRACE_EVENT1(ident, call, probe, arg1)

#define TRACE_EVENT2(ident, call, event, arg1, arg2) \
    KTR2(KTR_TRACE_FLAGS_##ident, \
         #call ":" #event \
         " " #arg1 "=0x%I64x" \
         " " #arg2 "=0x%I64x", \
         arg1, \
         arg2)
#define TRACE_ENTER2(ident, call, arg1, arg2) \
    TRACE_EVENT2(ident, call, enter, arg1, arg2)
#define TRACE_LEAVE2(ident, call, arg1, arg2) \
    TRACE_EVENT2(ident, call, leave, arg1, arg2)
#define TRACE_PROBE2(ident, call, arg1, arg2) \
    TRACE_EVENT2(ident, call, probe, arg1, arg2)

#define TRACE_EVENT3(ident, call, event, arg1, arg2, arg3) \
    KTR3(KTR_TRACE_FLAGS_##ident, \
         #call ":" #event \
         " " #arg1 "=0x%I64x" \
         " " #arg2 "=0x%I64x" \
         " " #arg3 "=0x%I64x", \
         arg1, \
         arg2, \
         arg3)
#define TRACE_ENTER3(ident, call, arg1, arg2, arg3) \
    TRACE_EVENT3(ident, call, enter, arg1, arg2, arg3)
#define TRACE_LEAVE3(ident, call, arg1, arg2, arg3) \
    TRACE_EVENT3(ident, call, leave, arg1, arg2, arg3)
#define TRACE_PROBE3(ident, call, arg1, arg2, arg3) \
    TRACE_EVENT3(ident, call, probe, arg1, arg2, arg3)

#define TRACE_EVENT4(ident, call, event, arg1, arg2, arg3, arg4) \
    KTR4(KTR_TRACE_FLAGS_##ident, \
         #call ":" #event \
         " " #arg1 "=0x%I64x" \
         " " #arg2 "=0x%I64x" \
         " " #arg3 "=0x%I64x" \
         " " #arg4 "=0x%I64x", \
         arg1, \
         arg2, \
         arg3, \
         arg4)
#define TRACE_ENTER4(ident, call, arg1, arg2, arg3, arg4) \
    TRACE_EVENT4(ident, call, enter, arg1, arg2, arg3, arg4)
#define TRACE_LEAVE4(ident, call, arg1, arg2, arg3, arg4) \
    TRACE_EVENT4(ident, call, leave, arg1, arg2, arg3, arg4)
#define TRACE_PROBE4(ident, call, arg1, arg2, arg3, arg4) \
    TRACE_EVENT4(ident, call, probe, arg1, arg2, arg3, arg4)


////////////////////////////////////////////////////////////////////////////

// Define exported variables here. (NOTE: Avoid exporting variables.)

// Ring buffer contains trace entries
//
TRACINGAPI KTR_ENTRY        g_TraceBuffer[KTR_ENTRY_MAX];
TRACINGAPI volatile LONG    g_TraceBufferSequence;
TRACINGAPI ULONG64          g_TraceMask;

////////////////////////////////////////////////////////////////////////////

// Define exported function prototypes here.
TRACINGAPI
void
LogTrace(
	KTR_TRACE_FLAGS flags,
	const CHAR*     Description,
	PVOID           Arguments1,
	PVOID           Arguments2,
	PVOID           Arguments3,
	PVOID           Arguments4,
	PVOID           Arguments5,
	PVOID           Arguments6,
	INT             NumOfArugments
);

////////////////////////////////////////////////////////////////////////////

// These are globals in the traced application

#ifdef THIS_IS_CLIENT

extern pfTrace pfLogTrace;
extern ULONG64* pTraceMask;

#undef THIS_IS_CLIENT
#endif // THIS_IS_CLIENT

#else // WIN32

#define KTR0(flags, Description)
#define KTR1(flags, Description, Arg1)
#define KTR2(flags, Description, Arg1, Arg2)
#define KTR3(flags, Description, Arg1, Arg2, Arg3)
#define KTR4(flags, Description, Arg1, Arg2, Arg3, Arg4)
#define KTR5(flags, Description, Arg1, Arg2, Arg3, Arg4, Arg5)
#define KTR6(flags, Description, Arg1, Arg2, Arg3, Arg4, Arg5, Arg6)

#endif // WIN32
#endif //UMTRACINGAPI_H